unit ConfigUtils;

{$mode objfpc}{$H+}

interface

uses
  SysUtils, Classes, Process, ContNrs;

type
  TCustomConfigTool = class;

  { TConfigToolList }

  TConfigToolList = class
  private
    FToolList: TObjectList;
    function GetCount: integer;

  public
    constructor Create;
    destructor Destroy; override;

    function Add(ConfigTool: TCustomConfigTool): TCustomConfigTool;
    function Run(index: integer): boolean;
    property Count: integer read GetCount;
  end;

  { TCustomConfigure }

  TCustomConfigure = class(TComponent)
  private
    FAppName: string;
    FSourcePath: string;
    FVersion: string;
    procedure SetAppName(AValue: string);
    procedure SetSourcePath(AValue: string);
    procedure SetVersion(AValue: string);
  public
    Tools: TConfigToolList;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function Run: boolean;

    property AppName: string read FAppName write SetAppName;
    property Version: string read FVersion write SetVersion;
    property SourcePath: string read FSourcePath write SetSourcePath;
  end;

  TFPCConfigure = class(TCustomConfigure)
  end;

  TConfigureClass = class of TCustomConfigure;

  { TCustomConfigTool }

  TCustomConfigTool = class(TObject)
  public
    function Run: boolean; virtual; abstract;
  end;

  { TFPIncFiles }

  TFPIncFiles = class(TCustomConfigTool)
  private
    FSourcePath: string;
    function FPMakeFileFind(search_path: string; file_list: TStrings): boolean;
    function FixProcName(AProcName: string): string;
    function CreateFPMakeIncFiles: boolean;
  public
    constructor Create(const ASourcePath: string);

    function Run: boolean; override;
    property SourcePath: string read FSourcePath;
  end;

  { TFPMakeBuild }

  TFPMakeBuild = class(TCustomConfigTool)
  private
    FSourcePath: string;
  public
    constructor Create(const ASourcePath: string);

    function Run: boolean; override;
    property SourcePath: string read FSourcePath;
  end;

function Configure(ConfigureClass: TConfigureClass): TCustomConfigure; overload;
function Configure: TCustomConfigure; overload;

implementation

var
  DefConfigure: TCustomConfigure;

function Configure(ConfigureClass: TConfigureClass): TCustomConfigure;
begin
  if not Assigned(DefConfigure) then
    DefConfigure := ConfigureClass.Create(nil);
  Result := DefConfigure;
end;

function Configure: TCustomConfigure;
begin
  Result := Configure(TFPCConfigure);
end;

procedure FPExecuteProcess(const CommandLine: string);
var
  AProcess: TProcess;
begin
  AProcess := TProcess.Create(nil);
  AProcess.CommandLine := CommandLine;
  AProcess.Options := AProcess.Options + [poWaitOnExit];
  AProcess.Execute;
  AProcess.Free;
end;

{ TFPMakeBuild }

constructor TFPMakeBuild.Create(const ASourcePath: string);
begin
  inherited Create;

  FSourcePath := ASourcePath;
end;

function TFPMakeBuild.Run: boolean;
begin
  Result := False;

  //build fpmake
  writeln('-- building fpmake file');
  FPExecuteProcess(Format('fpc %sfpmake.pp', [FSourcePath]));
  //FPExecuteProcess(Format('strip %sfpmake.exe', [FSourcePath]));
  //FPExecuteProcess(Format('upx -9 %sfpmake.exe', [FSourcePath]));

  //cleanup directory
  writeln('-- cleaning up');
  FPExecuteProcess('delp ' + FSourcePath);

  Result := True;
end;

{ TConfigToolList }

function TConfigToolList.GetCount: integer;
begin
  Result := FToolList.Count;
end;

constructor TConfigToolList.Create;
begin
  FToolList := TObjectList.Create(True);
end;

destructor TConfigToolList.Destroy;
begin
  FToolList.Free;

  inherited Destroy;
end;

function TConfigToolList.Add(ConfigTool: TCustomConfigTool): TCustomConfigTool;
begin
  FToolList.Add(ConfigTool);
  Result := ConfigTool;
end;

function TConfigToolList.Run(index: integer): boolean;
begin
  Result := TCustomConfigTool(FToolList[index]).Run;
end;

{ TFPIncFiles }

function TFPIncFiles.FPMakeFileFind(search_path: string; file_list: TStrings): boolean;
var
  err_code: integer;
  search_rec: TSearchRec;
  path_name: string;
  full_name: string;
  file_name: string;
  FileFind: boolean;
begin
  Result := False;

  path_name := ExtractFilePath(search_path);
  FileFind := DirectoryExists(path_name);

  if not FileFind then
    exit;

  file_name := ExtractFileName(search_path);
  err_code := FindFirst(search_path, faAnyFile, search_rec);

  while err_code = 0 do
  begin
    if search_rec.Name[1] <> '.' then
    begin
      full_name := path_name + search_rec.Name;

      if (search_rec.Attr and faDirectory > 0) then
        FPMakeFileFind(full_name + '\' + file_name, file_list)
      else
      if LowerCase(search_rec.Name) = 'fpmake.pp' then
        file_list.Add(full_name);
    end;
    err_code := FindNext(search_rec);
  end;

  Result := True;
end;

function TFPIncFiles.FixProcName(AProcName: string): string;
var
  i: integer;
begin
  Result := AProcName;

  for i := 0 to Length(AProcName) do
    if (Result[i] = '/') or (Result[i] = '\') then
      Result[i] := '_';
end;

function TFPIncFiles.CreateFPMakeIncFiles: boolean;
var
  file_list: TStrings;
  proc_file: TStrings;
  add_file: TStrings;
  i: integer;
  temp: string;
  folder_name: string;
begin
  Result := False;

  file_list := TStringList.Create;
  try
    FPMakeFileFind(FSourcePath + '*', file_list);

    proc_file := TStringList.Create;
    add_file := TStringList.Create;
    try
      for i := 0 to file_list.Count - 1 do
      begin
        temp := ExtractFilePath(file_list[i]);
        folder_name := copy(temp, Length(FSourcePath) + 1, Length(temp) - Length(FSourcePath) - 1);
        if folder_name <> '' then
        begin
          add_file.Add('  add_' + FixProcName(folder_name) + ';');

          proc_file.Add('procedure add_' + FixProcName(folder_name) + ';');
          proc_file.Add('begin');
          proc_file.Add('  with Installer do');
          proc_file.Add('    {$include ' + IncludeTrailingPathDelimiter(folder_name) + 'fpmake.pp}');
          proc_file.Add('end;');
        end;
      end;

      add_file.SaveToFile(FSourcePath + 'fpmake_add.inc');
      proc_file.SaveToFile(FSourcePath + 'fpmake_proc.inc');
    finally
      add_file.Free;
      proc_file.Free;
    end;

  finally
    file_list.Free;
  end;

  Result := True;
end;

constructor TFPIncFiles.Create(const ASourcePath: string);
begin
  inherited Create;

  FSourcePath := ASourcePath;
end;

function TFPIncFiles.Run: boolean;
begin
  Result := True;

  //create fpmake_add and fpmake_proc include files
  Write('-- creating fpmake include files');
  if CreateFPMakeIncFiles then
    writeln(' - done')
  else
  begin
    writeln(' - ERROR');
    halt(1);
  end;

  Result := False;
end;

{ TCustomConfigure }

procedure TCustomConfigure.SetVersion(AValue: string);
begin
  if FVersion = AValue then
    Exit;
  FVersion := AValue;
end;

constructor TCustomConfigure.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Tools := TConfigToolList.Create;
end;

destructor TCustomConfigure.Destroy;
begin
  Tools.Free;

  inherited Destroy;
end;

procedure TCustomConfigure.SetAppName(AValue: string);
begin
  if FAppName = AValue then
    Exit;
  FAppName := AValue;
end;

procedure TCustomConfigure.SetSourcePath(AValue: string);
begin
  if FSourcePath = AValue then
    Exit;
  FSourcePath := IncludeTrailingPathDelimiter(AValue);
end;

function TCustomConfigure.Run: boolean;
var
  i: integer;
begin
  Result := False;

  writeln(Format('-- configuring %s %s', [AppName, Version]));

  //basic tests

  //check source path
  if FSourcePath = '' then
  begin
    writeln('-- ERROR: source path not found ');
    halt(1);
  end
  else
    writeln('-- found source path: ', FSourcePath);

  //run all configure tools
  writeln;
  writeln('-- running configuration tools');
  for i := 0 to Tools.Count - 1 do
    Tools.Run(i);

  writeln;
  writeln('-- configuring done');

  Result := True;
end;

finalization
  FreeAndNil(DefConfigure);

end.
